/* -*- Mode: java; tab-width: 8; c-basic-offset: 4 -*-
 * Copyright  1997, 1998 Netscape Communications Corporation, All Rights
 * Reserved.
 */

package com.netscape.javascript.CORBA;

import com.netscape.javascript.*;
import java.util.Hashtable;
import org.omg.CORBA.DynamicImplementation;
import org.omg.CORBA.InterfaceDef;
import org.omg.CORBA.ServerRequest;
import org.omg.CORBA.Contained;
import org.omg.CORBA.OperationDef;
import org.omg.CORBA.OperationDefHelper;
import org.omg.CORBA.DefinitionKind;
import org.omg.CORBA.BAD_OPERATION;
import org.omg.CORBA.ParameterDescription;
import org.omg.CORBA.ParameterMode;
import org.omg.CORBA.NVList;
import org.omg.CORBA.Any;
import org.omg.CORBA.ARG_IN;
import org.omg.CORBA.IDLType;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.Bounds;
import org.omg.CORBA.CompletionStatus;

/**
 * This class reflects Scriptable objects into CORBA using the
 * CORBA Dynamic Skeleton Interface (DSI).
 *
 * @author Nick Thompson, Ruslan Belkin
 */

public class DynamicWrapper extends DynamicImplementation
{
    public DynamicWrapper(Scriptable obj, InterfaceDef idef, String name)
    {
        super (name, idef.describe_interface().id);     // always need to be called (ruslan)
        this.obj  =  obj;
        this.idef = idef;
        cx = new Context ();
    }

    public void invoke (ServerRequest request) 
    {
        String name = request.op_name();
        Contained[] contained;
    	OperationDef def = null;
        int i;

        try {
            cx.enter ();
        }   catch (Exception e) {};
        
	FlattenedObject  f = new FlattenedObject (obj);
	Function fun = getFunctionProperty (f, name);

	contained = idef.contents(DefinitionKind.dk_Operation, true);
	for (i = 0; i < contained.length; i++) 
	{
	    def = OperationDefHelper.narrow(contained[i]);
	    if (def.name().equals(name))
	        break;
	    def = null;
	}

	if (def == null)
	{
	    try {
                cx.exit ();
            }   catch (Exception e) {};
	    return;
	}

	ParameterDescription[] plist = def.params ();
	NVList params = _orb().create_operation_list (def);
	// fill in the parameters from the request
	request.params (params);

	// build the js argument list
	Object jsArgs[] = new Object[plist.length];

	for (i = 0; i < plist.length; i++) {
	    Any any;
	    try {
 	        any = params.item(i).value();
	    } 
	    catch (Bounds e) 
	    {
        	try {
                    cx.exit ();
                }   catch (Exception ee) {};
        	return;
	    }
	    jsArgs[i] = extractJSValue (any, plist[i].mode != ParameterMode.PARAM_IN);
	}

        // call function by name
	Object jsresult = fun.call (f.getObject (), jsArgs);

	// feed back any out parameters into params
	for (i = 0; i < plist.length; i++)
	    if (plist[i].mode != ParameterMode.PARAM_IN)
       	        try {
                    Marshal.convertIntoAny (jsArgs[i], params.item(i).value ());
                } 
	        catch (Bounds e)    { /* will never happen here */ }

        // throw exception (if any)
        Object except = Init.getThrownException ();

        if (except != null)
                request.except (((Struct)except).any);
        else    {                
        	// convert the result to the expected type
	        IDLType idlType = def.result_def ();
                request.result (makeAny(jsresult, idlType.type ()));
        }
        try {
            cx.exit ();
        }   catch (Exception ee) {};             
    }
    
    private Object extractJSValue (Any any, boolean isOut)
    {
        Object in = null;
        
        in = Marshal.convertFromAny (any);

	if (isOut) {
	    // create an object with one slot containing the value
	    in = new InOut (in);
	}
	return in;
    }

    private Any makeAny(Object jsval, TypeCode tc) 
    {
        // this is going to be a lot more complicated
	Any any = _orb().create_any();
	any.type (tc);

        Marshal.convertIntoAny (jsval, any);
        
	return any;
    }

    private Function getFunctionProperty (FlattenedObject f, String name) 
    {
        Object val = f.getProperty(name);
        if (val == null || !(val instanceof FlattenedObject))
            return null;
            
        Scriptable s = ((FlattenedObject) val).getObject();

        return s instanceof Function ? (Function) s : null;
    }

    Scriptable obj;
    InterfaceDef idef;

    Context cx;
